
/**************************************************************************
 *
 *  FILE            viking.c
 *  MODULE OF       
 *
 *  DESCRIPTION     
 *
 *  WRITTEN BY      Sverre H. Huseby
 *
 *  MODIFICATIONS   
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

#include <shhmsg.h>
#include <shhnet.h>
#include <xalloc.h>

#include "viking.h"

#define MAX_HOOKS 30

/**************************************************************************
 *                                                                        *
 *                       P R I V A T E    D A T A                         *
 *                                                                        *
 **************************************************************************/

typedef struct qe_tag {
    time_t when;
    char  *what;
    struct qe_tag *next;
} QueueEntry;
static QueueEntry head = { 0, 0, 0 };

typedef struct hook_tag {
    char *match;
    void (*f)(char *);
    int  report;
} HookEntry;
static HookEntry hook[MAX_HOOKS];
static int nHooks = 0;



/**************************************************************************
 *                                                                        *
 *                        P U B L I C    D A T A                          *
 *                                                                        *
 **************************************************************************/

int sock;  /* the socket referencing the connection to the Viking server */



/**************************************************************************
 *                                                                        *
 *                   P R I V A T E    F U N C T I O N S                   *
 *                                                                        *
 **************************************************************************/

static void checkOutQueue(void)
{
    QueueEntry *qe = head.next;
    time_t now;
    char *s;

    if (!qe)
        return;
    now = time(NULL);
    if (now >= qe->when) {
	s = xstrdup(qe->what);
        head.next = qe->next;
        free(qe->what);
        free(qe);
        vikOut(s);
	free(s);
    }
}

static void checkHooks(const char *line)
{
    int  q;
    char *s;
    
    for (q = 0; q < nHooks; q++)
        if (strstr(line, hook[q].match)) {
	    if (hook[q].report)
		msgVerbose(3, "***  calling hook-function for \"%s\"\n",
			   hook[q].match);
            s = xstrdup(line);
            hook[q].f(s);
            free(s);
        }
}

/* returns: -2 timeout, -1 error else character */
static int vikGetcTimeout(long mstimeout)
{
    SNErr err;
    unsigned char  c;
    struct timeval tv, *tvp;
    fd_set  fdset;
    int n;

    if (mstimeout >= 0) {
	tv.tv_sec = mstimeout / 1000L;
	tv.tv_usec = (mstimeout % 1000L) * 1000L;
	tvp = &tv;
    } else
	tvp = NULL;
    FD_ZERO(&fdset);
    FD_SET(sock, &fdset);
    n = FD_SETSIZE;
    if (snSelect(&n, &fdset, NULL, NULL, tvp) == SN_OK && n > 0) {
	if ((err = tcpReadAll(sock, &c, 1)) != SN_OK) {
	    msgVerbose(1, "*** tcpRead error on socket %d: %s\n",
		       sock, snErrToStr(err));
	    return -1;
	}
	return (int) c;
    }
    return -2;
}

static int vikGetsTimeout(char *s, int n, long mstimeout)
{
    int c, l = 0;

    for (;;) {
	c = vikGetcTimeout(mstimeout);
	if (c == -1)
	    return c;
	if (c == -2) {
	    --s;
	    if (l)
		c = 0;  /* prompt lines are not CR LF terminated */
	    break;
	}
	*s = c;
	++l;
	if (*s == '\n' || l == n - 1)
	    break;
	++s;
    }
    *(s + 1) = '\0';
    if (l && *s == '\n') {
	*s-- = '\0';
	--l;
    }
    if (l && *s == '\r')
	*s = '\0';
    return c < 0 ? c : 0;
}




/**************************************************************************
 *                                                                        *
 *                    P U B L I C    F U N C T I O N S                    *
 *                                                                        *
 **************************************************************************/

void vikSleep(int sec)
{
    char   line[200];
    struct timeval timeout;
    fd_set  fdset;
    int    n;

    msgVerbose(2, "***  sleeping for %d seconds\n", sec);
    timeout.tv_usec = 0;
    timeout.tv_sec = sec;
    while (timeout.tv_sec || timeout.tv_usec) {
	FD_ZERO(&fdset);
	FD_SET(sock, &fdset);
	n = FD_SETSIZE;
        if (snSelect(&n, &fdset, NULL, NULL, &timeout) == SN_OK && n > 0)
	    if (FD_ISSET(sock, &fdset)) {
		vikIn(line, sizeof(line));
	    }
    }
    msgVerbose(2, "***  done sleeping for %d seconds\n", sec);
}

void vikIn(char *line, int n)
{
#define MAX_TIMEOUTS 5
    int res, count;

    for (count = 0; count < MAX_TIMEOUTS; count++) {
	if ((res = vikGetsTimeout(line, n, 500L)) == -1)
	    msgFatalPerror("vikGetsTimeout");
	else if (res >= 0)
	    break;
	msgVerbose(1, "*** timeout %d in vikIn\n", count + 1);
    }
    if (count == MAX_TIMEOUTS)
	msgFatal("timeout in vikIn\n");
    msgVerbose(1, "In:  %s\n", line);
    checkHooks(line);
}

void vikOut(const char *format, ...)
{
    va_list ap;
    char buffer[200];

    checkOutQueue();
    va_start(ap, format);
    vsprintf(buffer, format, ap);
    va_end(ap);
    strcat(buffer, "\r\n");
    if (tcpWriteAll(sock, buffer, strlen(buffer)) != SN_OK)
	msgFatalPerror("tcpWriteAll");
    msgVerbose(1, "Out: %s", buffer);
}

void vikOutQueue(int delay, const char *format, ...)
{
    va_list ap;
    char buffer[200];
    QueueEntry *new, *qe;

    va_start(ap, format);
    vsprintf(buffer, format, ap);
    va_end(ap);
    new = xmalloc(sizeof(QueueEntry));
    new->when = time(NULL) + delay;
    new->what = xstrdup(buffer);
    qe = &head;
    while (qe->next) {
        if (qe->next->when >= new->when)
            break;
	qe = qe->next;
    }
    new->next = qe->next;
    qe->next = new;
}

int vikWaitForAndKeepTimeout(const char *match, char *line, int n,
			      long mstimeout)
{
    int  c, done = 0, l, lm, ret = 0;
    char *s;

    msgVerbose(2, "***  waiting for \"%s\"\n", match);
    lm = strlen(match);
    do {
	s = line;
	l = 0;
	for (;;) {
	    c = vikGetcTimeout(mstimeout);
	    if (c == -1)
		msgFatalPerror("tcpRead");
	    if (c == -2) {
		--s;
		ret = 1; /* timeout */
		msgVerbose(2, "***  timeout waiting for \"%s\"\n", match);
		goto finish;
	    }
	    *s = c;
	    ++l;
	    if (*s == '\n' || l == n - 1)
		break;
	    if (l >= lm && strncmp(s - lm + 1, match, lm) == 0) {
		done = 1;
		break;
	    }
	    ++s;
	}
	*(s + 1) = '\0';
	if (l && *s == '\n') {
	    *s-- = '\0';
	    --l;
	}
	if (l && *s == '\r')
	    *s = '\0';
	msgVerbose(1, "In:  %s\n", line);
        checkHooks(line);
    } while (!done);
    msgVerbose(2, "***  got \"%s\"\n", match);
  finish:
    return ret;
}

int vikWaitForTimeout(const char *match, long mstimeout)
{
    char buff[200];

    return vikWaitForAndKeepTimeout(match, buff, sizeof(buff), mstimeout);
}

void vikWaitForAndKeep(const char *match, char *line, int n)
{
    vikWaitForAndKeepTimeout(match, line, n, -1);
}

void vikWaitFor(const char *match)
{
    vikWaitForTimeout(match, -1);
}

void vikAddHook(const char *match, void (*f)(char *line), int report)
{
    if (nHooks == MAX_HOOKS)
        msgFatal("too many hooks\n");
    hook[nHooks].match = xstrdup(match);
    hook[nHooks].f = f;
    hook[nHooks].report = report;
    ++nHooks;
}

/* this function is not tested! */
void vikDelHook(const char *match, void (*f)(char *line))
{
    int q;

    for (q = 0; q < nHooks; )
        if (strcmp(hook[q].match, match) == 0 && hook[q].f == f) {
            free(hook[q].match);
            --nHooks;
            hook[q] = hook[nHooks];
        } else
            ++q;
}

void vikLogin(const char *name, const char *passwd)
{
    tcpClientOpen(HOST, PORT, &sock);
    sockSetLinger(sock, 1);
    vikWaitFor("name: ");
    vikOut("%s", name);
    vikWaitFor("assword: ");
    vikOut("%s", passwd);
}

void vikLogout(void)
{
    vikOut("save");
    vikOut("quit");
    sleep(2);
}

void vikLog(const char *filename, const char *format, ...)
{
    FILE *f;
    va_list ap;

    if ((f = fopen(filename, "at")) == NULL)
        msgFatal("cannot append to %s\n", filename);
    va_start(ap, format);
    vfprintf(f, format, ap);
    va_end(ap);
    fclose(f);
}
